@@ -116,21 +116,47 @@ KeyboardGrabber *Keyboard::currentGrab() const
     return m_grab;
 }
 
+void Keyboard::checkFocusResource(wl_keyboard::Resource *keyboardResource)
+{
+    if (!keyboardResource || !m_focus)
+        return;
+
+    // this is already the current  resource, do no send enter twice
+    if (m_focusResource == keyboardResource)
+        return;
+
+    // check if new wl_keyboard resource is from the client owning the focus surface
+    struct ::wl_client *focusedClient = m_focus->resource()->client();
+    if (focusedClient == keyboardResource->client()) {
+        sendEnter(m_focus, keyboardResource);
+        m_focusResource = keyboardResource;
+    }
+}
+
+void Keyboard::sendEnter(Surface *surface, wl_keyboard::Resource *keyboardResource)
+{
+    uint32_t serial = wl_display_next_serial(m_compositor->wl_display());
+    send_modifiers(keyboardResource->handle, serial, m_modsDepressed, m_modsLatched, m_modsLocked, m_group);
+    send_enter(keyboardResource->handle, serial, surface->resource()->handle, QByteArray::fromRawData((char *)m_keys.data(), m_keys.size() * sizeof(uint32_t)));
+}
+
 void Keyboard::focused(Surface *surface)
 {
-    if (m_focusResource && m_focus != surface) {
-        uint32_t serial = wl_display_next_serial(m_compositor->wl_display());
-        send_leave(m_focusResource->handle, serial, m_focus->resource()->handle);
+    if (m_focus != surface) {
+        if (m_focusResource) {
+            uint32_t serial = wl_display_next_serial(m_compositor->wl_display());
+            send_leave(m_focusResource->handle, serial, m_focus->resource()->handle);
+        }
         m_focusDestroyListener.reset();
+        if (surface) {
+            m_focusDestroyListener.listenForDestruction(surface->resource()->handle);
+        }
     }
 
     Resource *resource = surface ? resourceMap().value(surface->resource()->client()) : 0;
 
     if (resource && (m_focus != surface || m_focusResource != resource)) {
-        uint32_t serial = wl_display_next_serial(m_compositor->wl_display());
-        send_modifiers(resource->handle, serial, m_modsDepressed, m_modsLatched, m_modsLocked, m_group);
-        send_enter(resource->handle, serial, surface->resource()->handle, QByteArray::fromRawData((char *)m_keys.data(), m_keys.size() * sizeof(uint32_t)));
-        m_focusDestroyListener.listenForDestruction(surface->resource()->handle);
+        sendEnter(surface, resource);
     }
 
     m_focusResource = resource;
@@ -147,6 +173,13 @@ void Keyboard::setKeymap(const QWaylandKeymap &keymap)
 {
     m_keymap = keymap;
     m_pendingKeymap = true;
+
+    // If there is no key currently pressed, update right away the keymap
+    // Otherwise, delay the update when keys are released
+    // see http://lists.freedesktop.org/archives/wayland-devel/2013-October/011395.html
+    if (m_keys.isEmpty()) {
+        updateKeymap();
+    }
 }
 
 void Keyboard::focusDestroyed(void *data)
@@ -189,13 +222,17 @@ void Keyboard::keyboard_bind_resource(wl_keyboard::Resource *resource)
     if (m_context) {
         send_keymap(resource->handle, WL_KEYBOARD_KEYMAP_FORMAT_XKB_V1,
                     m_keymap_fd, m_keymap_size);
-        return;
     }
+    else
 #endif
-    int null_fd = open("/dev/null", O_RDONLY);
-    send_keymap(resource->handle, 0 /* WL_KEYBOARD_KEYMAP_FORMAT_NO_KEYMAP */,
-                null_fd, 0);
-    close(null_fd);
+    {
+        int null_fd = open("/dev/null", O_RDONLY);
+        send_keymap(resource->handle, 0 /* WL_KEYBOARD_KEYMAP_FORMAT_NO_KEYMAP */,
+                    null_fd, 0);
+        close(null_fd);
+    }
+
+    checkFocusResource(resource);
 }
 
 void Keyboard::keyboard_destroy_resource(wl_keyboard::Resource *resource)
@@ -216,17 +253,9 @@ void Keyboard::key(uint32_t serial, uint32_t time, uint32_t key, uint32_t state)
     }
 }
 
-void Keyboard::sendKeyEvent(uint code, uint32_t state)
+void Keyboard::keyEvent(uint code, uint32_t state)
 {
-    // There must be no keys pressed when changing the keymap,
-    // see http://lists.freedesktop.org/archives/wayland-devel/2013-October/011395.html
-    if (m_pendingKeymap && m_keys.isEmpty())
-        updateKeymap();
-
-    uint32_t time = m_compositor->currentTimeMsecs();
-    uint32_t serial = wl_display_next_serial(m_compositor->wl_display());
     uint key = code - 8;
-    m_grab->key(serial, time, key, state);
     if (state == WL_KEYBOARD_KEY_STATE_PRESSED) {
         m_keys << key;
     } else {
@@ -236,7 +265,14 @@ void Keyboard::sendKeyEvent(uint code, uint32_t state)
             }
         }
     }
-    updateModifierState(code, state);
+}
+
+void Keyboard::sendKeyEvent(uint code, uint32_t state)
+{
+    uint32_t time = m_compositor->currentTimeMsecs();
+    uint32_t serial = wl_display_next_serial(m_compositor->wl_display());
+    uint key = code - 8;
+    m_grab->key(serial, time, key, state);
 }
 
 void Keyboard::modifiers(uint32_t serial, uint32_t mods_depressed,
@@ -280,6 +316,11 @@ void Keyboard::updateModifierState(uint code, uint32_t state)
 
 void Keyboard::updateKeymap()
 {
+    // There must be no keys pressed when changing the keymap,
+    // see http://lists.freedesktop.org/archives/wayland-devel/2013-October/011395.html
+    if (!m_pendingKeymap || !m_keys.isEmpty())
+        return;
+
     m_pendingKeymap = false;
 #ifndef QT_NO_WAYLAND_XKB
     if (!m_context)
